#!/usr/bin/perl -w

# probepicker7.pl
# Purpose: The design of highly unique oligomers for a given sequence
# Input: Concatenated FASTA file of sequences for which to design probes
# Output: Tab-delimited text file of probes and characteristics
# Dependencies: Requires installed WU-BLAST 2.0 and PRIMER3, with
#               appropriate paths defined.  Also requires a BLAST
#               database against which to search.
# Written by Charlie Kim
# Algorithm developed by Yves Goguet and Charlie Kim
# Written March 30, 2001
# Copyright 2001

# Modified June 12, 2001 to include oligo Tm matching, better error handling,
#         more data reporting, new BLAST score calculation

###########
# MODULES #
###########
use Bio::SeqIO;

#############
# VARIABLES #
#############

$infile = "test.seq";
$outfile = "test.probes";
$tempblsfile = "temporaryblastout";
$tempoligofile = "temporaryoligofile";
$errorfile = "test.errors";
$blastprogram = "blastn"; # the full path should be specified if it is not
                          # included in your environment variables
$blastdb = "testdb";
$blastscorethreshold = "S=10"; # can also use "E=#" or "" for default settings
$blastword = "W=11"; # default = 11 for blastn, use smaller for more hits
$blastseqreport = "noseqs"; # values: "noseqs"(no sequences)  or ""(sequences)
$blastgap = ""; # values: "nogap"(no gaps) or ""(gaps)
$oligos_to_save = 10; # the number of oligo probes to return for each sequence
@scoreweightmatrix = (4,3,2,1); # skips the perfect hit, then gives
                                # a relative weight of 4 to the 2nd hit,
                                # 3 to the 3rd hit, 2 to the 4th hit,
                                # 1 to the 5th hit, etc.
                                # The entire matrix can be shifted to 
                                # include additional scores, and the
                                # weights may also be modified
$tempprimer3file = "temporaryprimer3file";
$minlength = 26; # minimum probe length
$maxlength = 30; # maximum probe length
$method = 1; # method 1 = 2+4, method 2 = howley et al (0.41*GC)
$optimalPercentGC = 0.65614; # average %GC of M. tb
$salt_conc = 0.9; # for use in Tm calculations

#########
# TESTS #
#########

my $flag = 0;
open(PRIMERTEST, "which primer3_core |") or die "Failed to test for primer3_core\n";
while(<PRIMERTEST>) {
    $flag++;
}
close PRIMERTEST;
die "Unable to execute primer3_core\n" unless $flag;

$flag = 0;
open(BLSTEST, "which $blastprogram |") or die "Failed to test for primer3_core\n";
while(<BLSTEST>) {
    $flag++;
}
close BLSTEST;
die "Unable to execute $blastprogram\n" unless $flag;


########
# MAIN #
########

$HSPkeeps = $#scoreweightmatrix + 2;

open(OUTFILE, ">$outfile") or die "writeopenfail: $outfile\n";
open(ERRORFILE, ">$errorfile") or die "writeopenfail: $errorfile\n";
print OUTFILE "Query\t";
for ($q = 1; $q <= $oligos_to_save; $q++) {
    print OUTFILE "OligoNumber$q\tOligoSequence$q\tOligoLength$q\tMethod$method","Tm$q\tOligoScore$q\tPenalty$q\tTm$q\tG\+C$q\tSelfAnneal$q\tSelfComplement$q\t";
}
print OUTFILE "\n";


my $seqfile = Bio::SeqIO->new(-file => $infile,
			      -format => 'fasta');

while (defined(my $query = $seqfile->next_seq)) {
    $gene = $query->id;
    $gene = '>' . $gene;
    print "$gene\n";
    $seq = $query->seq;
#    print "$seq\n";

    if (length($seq) < $maxlength) {
	print ERRORFILE "Sequence shorter than oligo length: $gene\n";
	next;
    }
    @output = &getbestoligos($gene, $seq);
    if (!@output) {
	print ERRORFILE "No oligos found for $gene\n";
	next;
	}
    $gene =~ s/^>//;
    print OUTFILE "$gene\t";
    foreach $entry (@output) {
	print OUTFILE "$entry\t";
    }
    print OUTFILE "\n";
}
#    }
#}


###############
# SUBROUTINES #
###############

sub getbestoligos {
    my @seq = split('',$_[1]);
    my $query = '';
    my $oligo = '';
    my %oligoseqs = ();
    my %oligoscores = ();
    my %calctemp = ();
    my %primer3stats = ();
    my @lines = ();

    for ($c = 0; $c + $maxlength <= $#seq + 1; $c++) {
	$| = 1;
	print "\b" x 30;
	print "Analyzing oligo $c";
	$query = "$_[0] (Oligo $c)";
	$fulloligo = '';
	for ($d = 0; $d < $maxlength; $d++) { # these lines will
	    $fulloligo .= $seq[$c+$d];              # be replaced with
	}                                       # a getoligoseq sub
	($oligo,$calctemp{$query}) = &matchOligoTm($fulloligo,$method,$minlength,$maxlength,$optimalPercentGC);
	$oligolength = length($oligo);
	my $weightedscore = &runblast($query, $oligo, $oligolength);
	if (!$weightedscore) {
	    print ERRORFILE "No perfect hit for $query\n";
	    next;
	}
	$oligoseqs{$query} = $oligo;
	$oligoscores{$query} = $weightedscore;
    }
    print "\n";
    if (!%oligoscores) {
	return;
    }
    my @topkeys = &gettopscores(%oligoscores);
    print "Analyzing top oligo characteristics\n";
    foreach $value (@topkeys) {
	$primer3stats{$value} = &primer3stats("$oligoseqs{$value}");
	$value =~ /\(Oligo (\d*)\)/;
	$oligonumber = $1;
	$oligolength = length($oligoseqs{$value});
	push @lines, "$oligonumber\t$oligoseqs{$value}\t$oligolength\t$calctemp{$value}\t$oligoscores{$value}\t$primer3stats{$value}"; 
    }
    return @lines;
}

sub matchOligoTm {
    my ($s,$meth,$min,$max,$optgc) = @_;
    my @length = ();
    my @gc_num = ();
    my %gc_pct = ();
    @seq = split('',$s);
    my $gc_num = 0;
    for (my $c = 0; $c < $min; $c++) {
	$gc_num++ if $seq[$c] =~ /[G|C]/i;
    }
    $gc_num[0] = $gc_num;
    $length[0] = $min;
    $gc_pct{$min} = $gc_num[0] / $length[0];
    for (my $d = 1; $d <= $max-$min; $d++) {
	$gc_num++ if $seq[$min+$d-1] =~ /[G|C]/i;
	$gc_num[$d] = $gc_num;
	$length[$d] = $min+$d;
	$gc_pct{$length[$d]} = $gc_num[$d] / $length[$d];
    }

# MATCH BY GC COUNT #
    if ($meth == 1) {
	my %scores = ();
	my $gc_bases = $optgc * ($max+$min)/2;
	my $at_bases = ($max+$min)/2 - $gc_bases;
	my $optTm = (2 * $at_bases) + (4 * $gc_bases);
	my $bestscore = 1000; # arbitrary high number
	for (my $e = 0; $e <= $#gc_num; $e++) {
	    my $temp = 2*($length[$e]-$gc_num[$e]) + 4*$gc_num[$e];
	    $score{$length[$e]} = abs($temp-$optTm);
	    if ($bestscore > $score{$length[$e]}) {
		$bestscore = $score{$length[$e]}; # can be removed
		$besthit = $length[$e];
		$besttemp = $temp;
	    }
	}
    }

# MATCH BY HOWLEY #
    elsif ($meth == 2) {
	my %scores = ();
	my $optTm = 81.5 + 41*$optgc - 1000/($max+$min) + 16.6*log($salt_conc)/log(10);
	my $bestscore = 1000;
	for (my $e = 0; $e <= $#gc_num; $e++) {
	    my $temp =  81.5 + 41*$gc_pct{$length[$e]} - 500/$length[$e] + 16.6*log($salt_conc)/log(10);
	    $score{$length[$e]} = abs($temp-$optTm);
	    if ($bestscore > $score{$length[$e]}) {
		$bestscore = $score{$length[$e]}; # can be removed
		$besthit = $length[$e];
		$besttemp = $temp;
	    }
	}
    }

# end of methods #
    else { die "Method not properly specified\n"; }
    my $matchedoligo = '';
    for (my $f = 0; $f < $besthit; $f++) {
	$matchedoligo .= $seq[$f];
    }
    return ($matchedoligo, $besttemp); # remove besttemp if removed above
}

sub runblast {
    my @HSPscores;
    &createoligofile($_[0], $_[1]);
    my $oligolength = $_[2];
    &createblastfile;
    my $weightedscore = &getbestHSPscores($oligolength);
    unlink "$tempoligofile";
    unlink "$tempblsfile";
    return $weightedscore;
}


sub createoligofile {
    open(OLIGOFILE, ">$tempoligofile") or die "writeopenfail: $tempoligofile\n";
    print OLIGOFILE "$_[0]\n$_[1]\n";
    close OLIGOFILE;
}

sub createblastfile {
    open(BLASTRUN, "$blastprogram $blastdb temporaryoligofile $blastseqreport $blastgap $blastword $blastscorethreshold \-hspmax $HSPkeeps \-warnings|") or die "programfail: Can't open BLAST\n";
    open(BLSFILE, ">$tempblsfile") or die "writeopenfail: $tempblsfile\n";
    while(<BLASTRUN>) {
	print BLSFILE $_;
    }
    close BLSFILE;
}

sub getbestHSPscores {
    my $strandflag = '';
    my @allscores = ();
    my $oligolength = $_[0];

    open(BLSFILE, "$tempblsfile") or die "readopenfail: $tempblsfile\n";
    while(<BLSFILE>) {
	print ERRORFILE "BLAST file read terminated!  No hits found for\n$gene\n\n" if (/\*\*\* NONE \*\*\*/);
	if (/Identities \= (\d*)\/(\d*)/) {
	    my $score = ($1 * $1) / ($2 * $oligolength); # score calculation
	    push @allscores, $score;
	}
    }
    if ($#allscores < $#scoreweightmatrix) {
	print "Not enough HSP scores obtained: try a lower stringency score for the BLAST\n";
	$position = $#allscores+1;
	while ($position <= $#scoreweightmatrix+1) {
	    $allscores[$position] = 0.000001;
	    $position++;
	}
    }
    my $weightedscore = &calculatescores(@allscores);
    return $weightedscore;
}
	
sub calculatescores {
    my @scores = sort { $b <=> $a } @_;
    my $sumscore = 0;
    my $totalweight = 0;
    my $weightedscore = 0;
    if ($scores[0] != 1) { # must be modified for score calculation
#	print ERRORFILE "No perfect hit for $gene\n";
	return;
    }
    for (my $g = 0; $g <= $#scoreweightmatrix; $g++) {
	$sumscore += $scoreweightmatrix[$g] * $scores[$g+1];
	$totalweight += $scoreweightmatrix[$g];
    }
    $weightedscore = $sumscore / $totalweight;
    return $weightedscore;
}


sub gettopscores {
    my %oligoscores = @_;
    my @topkeys = ();
    my $counter = 0;
    my @sortedscores = sort { $oligoscores{$a} <=> $oligoscores{$b} } keys %oligoscores;
    while ($counter < $oligos_to_save) {
	push @topkeys, $sortedscores[$counter];
	$counter++;
    }
    return @topkeys;
}

sub primer3stats {
    &primer3makefile($_[0]);
    open(PRIMERRUN, "primer3_core < $tempprimer3file |") or die "Can't run Primer3\n";
    while (<PRIMERRUN>) {
#	$seq_id = "$1\t" if (/PRIMER_SEQUENCE_ID\=(.*)/ );
#	$oligo_seq = "$1\t" if (/PRIMER_INTERNAL_OLIGO_SEQUENCE\=(.*)/ );
	$penalty = "$1" if (/PRIMER_INTERNAL_OLIGO_PENALTY\=(.*)/ );
	$Tm = "$1" if (/PRIMER_INTERNAL_OLIGO_TM\=(.*)/ );
	$GC = "$1" if (/PRIMER_INTERNAL_OLIGO_GC_PERCENT\=(.*)/ );
	$self_any = "$1" if (/PRIMER_INTERNAL_OLIGO_SELF_ANY\=(.*)/ );
	$self_end = "$1" if (/PRIMER_INTERNAL_OLIGO_SELF_END\=(.*)/ );
	$error = "$1" if (/PRIMER_ERROR\=(.*)/ );
    }
    unlink "$tempprimer3file";
    if ($error) {
	return ("$error\t\t\t\t");
    }
    else {
	return ("$penalty\t$Tm\t$GC\t$self_any\t$self_end");
    }
}

sub primer3makefile {
    my @parameters = ( "PRIMER_TASK=pick_hyb_probe_only",
		       "PRIMER_INTERNAL_OLIGO_OPT_SIZE=$oligolength",
		       "PRIMER_INTERNAL_OLIGO_MIN_SIZE=$oligolength",
		       "PRIMER_INTERNAL_OLIGO_MAX_SIZE=$oligolength",
		       "PRIMER_INTERNAL_OLIGO_OPT_TM=75",
		       "PRIMER_INTERNAL_OLIGO_MIN_TM=10",
		       "PRIMER_INTERNAL_OLIGO_MAX_TM=200",
		       );
    push @parameters, "SEQUENCE=$_[0]\n=";
    open(PRIMER3FILE, ">$tempprimer3file") or die "writefilefail: $tempprimer3file\n";
    foreach (@parameters) {
	print PRIMER3FILE "$_\n";
    }
}
